import greenfoot.*;  // (Actor, World, Greenfoot, GreenfootImage).
import java.util.*; // In particular, this import will allow me to utilize lists.
import java.lang.*; // In particular, this import will allow me to use Intger.parseInt(java.lang.string);

/**
 * This class defines an Animal. Animals have a fized size and their color can either 
 * depend on their parents or a mutation. Animals live in the Wilderness.
 * 
 * @author Isaac Chen 
 * @version Version 2, February 2023
 */
public class Animal extends Actor
{
    private int vegetationEaten;
    private int daysLeft;
    
    private int redPart;
    private int greenPart;
    private int bluePart;
        
    /**
     * Default constructor of an Animal. Makes the object's color
     * dark blue and initializes the other variables.
     */
    public Animal() 
    {   
        this((new Color(0, 0, 139)));
    }
    
    /**
     * Constructs an Animal with a specified color, using 
     * the color parameter.
     * All Animals will be circles of diameter 15 pixels.
     */
    public Animal(Color color)
    {
        redPart = color.getRed();
        greenPart = color.getGreen();
        bluePart = color.getBlue();
        
        GreenfootImage image = new GreenfootImage(15, 15);
        image.setColor(color);
        image.fillOval(0, 0, 15, 15); 
        setImage(image);
        
        vegetationEaten = 0;
        // Wilderness.currentPopulation++;
        // This line is problematic becaue it messes up the inital value of currentPopulation. 
        
        daysLeft = Greenfoot.getRandomNumber(100) + Wilderness.scaleUp;
        // Range of lifespan is determined by user input from the Wilderness class.
    }
    
    /**
     * First checks if food can be eaten. 
     * Next, checks if the current object is overlapping with 
     * another Animal and runs Greenfoot.getRandomNumber to 
     * see if an offspring (which may have a mutation) will be initiated. Updates 
     * the relevant variables.
     * Then, either moves and turns or just turns.
     * Lastly, variables are updated and the checkDeath method is 
     * run (which has the potential to end the current object's 
     * life by removing it).
     */
    public void act()
    {
        if (foundVegetation()) 
        {
            eatVegetation();
        }
        
        if(foundAnotherAnimal())
        {
            if(Greenfoot.getRandomNumber(100) < 25)
            {                
                if(Greenfoot.getRandomNumber(20) == 0)
                {
                    int r = Greenfoot.getRandomNumber(255);
                    int g = Greenfoot.getRandomNumber(255);
                    int b = Greenfoot.getRandomNumber(255);
                    
                    int startX = Greenfoot.getRandomNumber(getWorld().getWidth());
                    int startY = Greenfoot.getRandomNumber(getWorld().getHeight());
                    
                    getWorld().addObject(new Animal(new Color(r, g, b)), startX, startY);
                }
                
                else
                {   
                    averageColorValuesAndAddChild();
                }
                
                Wilderness.currentPopulation++;
            }
        }
        
        if (canMove()) 
        {
            randMove();
            turnRandom();
        }
        else
        {
            turnRandom();
        }
        
        daysLeft--;
        
        checkDeath();
    }   
    
    /**
     * This method checks if one of the conditions necessary for death 
     * is attained. If so, then the current object is removed and the
     * Wilderness static variable is decreased by 1.
     * 
     * Note: The purpose of the 'else if' condition is to limit the 
     * current population of the world. It mimics the concept of a 
     * "carrying capacity" in Evolutionary Biology. 
     */
    private void checkDeath()
    {
       if(daysLeft == 0)
       {
           Wilderness.currentPopulation--;
           getWorld().removeObject(this);
       }
    
       else if(Wilderness.currentPopulation > 200 && Greenfoot.getRandomNumber(5) < 2)
       {
           Wilderness.currentPopulation--;
           getWorld().removeObject(this);
       }
    }
    
    /**
     * When this method is run, the current object moves forward by 
     * n \in [10, 19] pixels with equal probability. 
     */
    public void randMove()
    {
        int randNumMoves = Greenfoot.getRandomNumber(10) + 10;
        move(randNumMoves);
    }
    
    /**
     * Turns the current object by a random number of degrees.
     * Equivalent to picking a random direction for the current
     * object each app cycle.
     */
    public void turnRandom()
    {
        int direction = Greenfoot.getRandomNumber(360);
        turn(direction);
    }
    
    /**
     * Checks if the current Animal is overlapping with another Animal.
     * Returns true or false.
     */
    public boolean foundAnotherAnimal()
    {
        Actor otherAnimal = getOneObjectAtOffset(0, 0, Animal.class);
        return otherAnimal != null;
    }
    
    /**
     * Called when an overlap of Animals results in the production of 
     * a non-mutated offspring initiated by the act method of the 
     * current object.
     * Note: See if(foundAnotherAnimal()) from the act method.
     * 
     * Gets the overlapping Actor from the Animal class, casts it 
     * as an Animal object, and finally averages the RGB values of 
     * the parents and adds the child with said RGB values at a random 
     * location within the world.
     */
    private void averageColorValuesAndAddChild()
    {
        Actor otherAnimal = getOneObjectAtOffset(0, 0, Animal.class);
        Animal theMate = (Animal) otherAnimal; 
        /**
         * Here, we are casting otherAnimal into an Animal Object so 
         * that we can access its RGB values, as they are instance 
         * variables from the Animal class.
         */
        
        int avgR = (redPart + theMate.redPart)/2;
        int avgG = (greenPart + theMate.greenPart)/2;
        int avgB = (bluePart + theMate.bluePart)/2;
        
        int startX = Greenfoot.getRandomNumber(getWorld().getWidth());
        int startY = Greenfoot.getRandomNumber(getWorld().getHeight());
        
        getWorld().addObject(new Animal(new Color(avgR, avgG, avgB)), startX, startY);
    }
    
    /**
     * Checks if the current Animal is overlapping with a bush.
     * Returns true or false.
     */
    public boolean foundVegetation() 
    {
        Actor crab = getOneObjectAtOffset(0, 0, Vegetation.class);
        return crab != null;
    } 
   
   /**
     * Checks if the current Animal is overlapping with a bush. If so,
     * then the bush is removed (eaten) and the relevant variables are 
     * incremented.
     */
    public void eatVegetation()
    {
        Actor vegetation = getOneObjectAtOffset(0, 0, Vegetation.class);
        if (vegetation != null) 
        {
            getWorld().removeObject(vegetation);
            // Here, removing the vegetation is essentially eating the vegetation.
            vegetationEaten++; 
            Wilderness.totalVegetationEaten++;
        }
    }
    
    /**
     * Checks if the current object can move forward. 
     * Returns true if we can, false otherwise.
     */
    public boolean canMove()
    {
        int rotation = getRotation();
        int x = getX();
        int y = getY();
        boolean facingEdge = false;
        
        switch (rotation) {
            case 0:
                facingEdge = (x == getWorld().getWidth() - 1);
                break;
            case 90:
                facingEdge = (y == getWorld().getHeight() - 1);
                break;
            case 180:
                facingEdge = (x == 0);
                break;
            case 270:
                facingEdge = (y == 0);
                break;
        }
        
        return !facingEdge;
    }

    /**
     * Returns how many bushes the current object has eaten.
     */
    public int getVegetationEaten()
    {
        return vegetationEaten;
    }
}